@@ -4,25 +4,26 @@ import collections.abc
 import inspect
 import sys
 import types
+import typing
 import warnings
+from collections.abc import Mapping, MutableMapping, Sequence
 from enum import Enum
 from inspect import Parameter, isclass, isfunction
 from io import BufferedIOBase, IOBase, RawIOBase, TextIOBase
+from itertools import zip_longest
 from textwrap import indent
 from typing import (
     IO,
     AbstractSet,
+    Annotated,
     Any,
     BinaryIO,
     Callable,
     Dict,
     ForwardRef,
     List,
-    Mapping,
-    MutableMapping,
     NewType,
     Optional,
-    Sequence,
     Set,
     TextIO,
     Tuple,
@@ -32,40 +33,43 @@ from typing import (
 )
 from unittest.mock import Mock
 
-from ._config import ForwardRefPolicy, global_config
+import typing_extensions
+
+# Must use this because typing.is_typeddict does not recognize
+# TypedDict from typing_extensions, and as of version 4.12.0
+# typing_extensions.TypedDict is different from typing.TypedDict
+# on all versions.
+from typing_extensions import is_typeddict
+
+from ._config import ForwardRefPolicy
 from ._exceptions import TypeCheckError, TypeHintWarning
-from ._memo import CallMemo, TypeCheckMemo
-from ._utils import evaluate_forwardref, get_type_name, qualified_name
+from ._memo import TypeCheckMemo
+from ._utils import evaluate_forwardref, get_stacklevel, get_type_name, qualified_name
+
+if sys.version_info >= (3, 15):
+    from typing import NoExtraItems
+else:
+    from typing_extensions import NoExtraItems
 
 if sys.version_info >= (3, 11):
     from typing import (
-        Annotated,
-        Literal,
-        LiteralString,
-        Self,
+        NotRequired,
+        Required,
         TypeAlias,
-        TypeGuard,
         get_args,
         get_origin,
-        get_type_hints,
-        is_typeddict,
     )
 
     SubclassableAny = Any
 else:
+    from typing_extensions import Any as SubclassableAny
     from typing_extensions import (
-        Annotated,
-        Literal,
-        LiteralString,
-        Self,
+        NotRequired,
+        Required,
         TypeAlias,
-        TypeGuard,
         get_args,
         get_origin,
-        get_type_hints,
-        is_typeddict,
     )
-    from typing_extensions import Any as SubclassableAny
 
 if sys.version_info >= (3, 10):
     from importlib.metadata import entry_points
@@ -82,7 +86,11 @@ TypeCheckLookupCallback: TypeAlias = Callable[
 ]
 
 checker_lookup_functions: list[TypeCheckLookupCallback] = []
-
+generic_alias_types: tuple[type, ...] = (
+    type(List),
+    type(List[Any]),
+    types.GenericAlias,
+)
 
 # Sentinel
 _missing = object()
@@ -143,7 +151,10 @@ BINARY_MAGIC_METHODS = {
 
 
 def check_callable(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if not callable(value):
         raise TypeCheckError("is not callable")
@@ -155,7 +166,9 @@ def check_callable(
             return
 
         argument_types = args[0]
-        if argument_types is not Ellipsis:
+        if isinstance(argument_types, list) and not any(
+            type(item) is ParamSpec for item in argument_types
+        ):
             # The callable must not have keyword-only arguments without defaults
             unfulfilled_kwonlyargs = [
                 param.name
@@ -166,40 +179,41 @@ def check_callable(
             if unfulfilled_kwonlyargs:
                 raise TypeCheckError(
                     f"has mandatory keyword-only arguments in its declaration: "
-                    f'{", ".join(unfulfilled_kwonlyargs)}'
+                    f"{', '.join(unfulfilled_kwonlyargs)}"
                 )
 
-            num_mandatory_args = len(
-                [
-                    param.name
-                    for param in signature.parameters.values()
-                    if param.kind
-                    in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
-                    and param.default is Parameter.empty
-                ]
-            )
-            has_varargs = any(
-                param
-                for param in signature.parameters.values()
-                if param.kind == Parameter.VAR_POSITIONAL
-            )
+            num_positional_args = num_mandatory_pos_args = 0
+            has_varargs = False
+            for param in signature.parameters.values():
+                if param.kind in (
+                    Parameter.POSITIONAL_ONLY,
+                    Parameter.POSITIONAL_OR_KEYWORD,
+                ):
+                    num_positional_args += 1
+                    if param.default is Parameter.empty:
+                        num_mandatory_pos_args += 1
+                elif param.kind == Parameter.VAR_POSITIONAL:
+                    has_varargs = True
 
-            if num_mandatory_args > len(argument_types):
+            if num_mandatory_pos_args > len(argument_types):
                 raise TypeCheckError(
-                    f"has too many arguments in its declaration; expected "
-                    f"{len(argument_types)} but {num_mandatory_args} argument(s) "
-                    f"declared"
+                    f"has too many mandatory positional arguments in its declaration; "
+                    f"expected {len(argument_types)} but {num_mandatory_pos_args} "
+                    f"mandatory positional argument(s) declared"
                 )
-            elif not has_varargs and num_mandatory_args < len(argument_types):
+            elif not has_varargs and num_positional_args < len(argument_types):
                 raise TypeCheckError(
                     f"has too few arguments in its declaration; expected "
-                    f"{len(argument_types)} but {num_mandatory_args} argument(s) "
+                    f"{len(argument_types)} but {num_positional_args} argument(s) "
                     f"declared"
                 )
 
 
 def check_mapping(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if origin_type is Dict or origin_type is dict:
         if not isinstance(value, dict):
@@ -213,7 +227,7 @@ def check_mapping(
     if args:
         key_type, value_type = args
         if key_type is not Any or value_type is not Any:
-            samples = global_config.collection_check_strategy.iterate_samples(
+            samples = memo.config.collection_check_strategy.iterate_samples(
                 value.items()
             )
             for k, v in samples:
@@ -231,26 +245,54 @@ def check_mapping(
 
 
 def check_typed_dict(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
+    if not isinstance(value, dict):
+        raise TypeCheckError("is not a dict")
+
     declared_keys = frozenset(origin_type.__annotations__)
-    if hasattr(origin_type, "__required_keys__"):
-        required_keys = origin_type.__required_keys__
-    else:  # py3.8 and lower
-        required_keys = declared_keys if origin_type.__total__ else frozenset()
+    required_keys = set(origin_type.__required_keys__)
+    existing_keys = set(value)
+    if extra_keys := existing_keys - declared_keys:
+        if (
+            argtype := getattr(origin_type, "__extra_items__", NoExtraItems)
+        ) is NoExtraItems:
+            keys_formatted = ", ".join(
+                f'"{key}"' for key in sorted(extra_keys, key=repr)
+            )
+            raise TypeCheckError(f"has unexpected extra key(s): {keys_formatted}")
 
-    existing_keys = frozenset(value)
-    extra_keys = existing_keys - declared_keys
-    if extra_keys:
-        keys_formatted = ", ".join(f'"{key}"' for key in sorted(extra_keys, key=repr))
-        raise TypeCheckError(f"has unexpected extra key(s): {keys_formatted}")
+        for key in extra_keys:
+            argvalue = value[key]
+            try:
+                check_type_internal(argvalue, argtype, memo)
+            except TypeCheckError as exc:
+                exc.append_path_element(f"value of key {key!r}")
+                raise
 
-    missing_keys = required_keys - existing_keys
-    if missing_keys:
+    # Detect NotRequired fields which are hidden by get_type_hints()
+    type_hints: dict[str, type] = {}
+    for key, annotation in origin_type.__annotations__.items():
+        if isinstance(annotation, ForwardRef):
+            annotation = evaluate_forwardref(annotation, memo)
+
+        if get_origin(annotation) is NotRequired:
+            required_keys.discard(key)
+            annotation = get_args(annotation)[0]
+        elif get_origin(annotation) is Required:
+            required_keys.add(key)
+            annotation = get_args(annotation)[0]
+
+        type_hints[key] = annotation
+
+    if missing_keys := required_keys - existing_keys:
         keys_formatted = ", ".join(f'"{key}"' for key in sorted(missing_keys, key=repr))
         raise TypeCheckError(f"is missing required key(s): {keys_formatted}")
 
-    for key, argtype in get_type_hints(origin_type).items():
+    for key, argtype in type_hints.items():
         argvalue = value.get(key, _missing)
         if argvalue is not _missing:
             try:
@@ -261,13 +303,16 @@ def check_typed_dict(
 
 
 def check_list(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if not isinstance(value, list):
         raise TypeCheckError("is not a list")
 
     if args and args != (Any,):
-        samples = global_config.collection_check_strategy.iterate_samples(value)
+        samples = memo.config.collection_check_strategy.iterate_samples(value)
         for i, v in enumerate(samples):
             try:
                 check_type_internal(v, args[0], memo)
@@ -277,13 +322,16 @@ def check_list(
 
 
 def check_sequence(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if not isinstance(value, collections.abc.Sequence):
         raise TypeCheckError("is not a sequence")
 
     if args and args != (Any,):
-        samples = global_config.collection_check_strategy.iterate_samples(value)
+        samples = memo.config.collection_check_strategy.iterate_samples(value)
         for i, v in enumerate(samples):
             try:
                 check_type_internal(v, args[0], memo)
@@ -293,13 +341,19 @@ def check_sequence(
 
 
 def check_set(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
-    if not isinstance(value, AbstractSet):
+    if origin_type is frozenset:
+        if not isinstance(value, frozenset):
+            raise TypeCheckError("is not a frozenset")
+    elif not isinstance(value, AbstractSet):
         raise TypeCheckError("is not a set")
 
     if args and args != (Any,):
-        samples = global_config.collection_check_strategy.iterate_samples(value)
+        samples = memo.config.collection_check_strategy.iterate_samples(value)
         for v in samples:
             try:
                 check_type_internal(v, args[0], memo)
@@ -309,14 +363,13 @@ def check_set(
 
 
 def check_tuple(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     # Specialized check for NamedTuples
-    field_types = getattr(origin_type, "__annotations__", None)
-    if field_types is None and sys.version_info < (3, 8):
-        field_types = getattr(origin_type, "_field_types", None)
-
-    if field_types:
+    if field_types := getattr(origin_type, "__annotations__", None):
         if not isinstance(value, origin_type):
             raise TypeCheckError(
                 f"is not a named tuple of type {qualified_name(origin_type)}"
@@ -334,7 +387,6 @@ def check_tuple(
         raise TypeCheckError("is not a tuple")
 
     if args:
-        # Python 3.6+
         use_ellipsis = args[-1] is Ellipsis
         tuple_params = args[: -1 if use_ellipsis else None]
     else:
@@ -343,7 +395,7 @@ def check_tuple(
 
     if use_ellipsis:
         element_type = tuple_params[0]
-        samples = global_config.collection_check_strategy.iterate_samples(value)
+        samples = memo.config.collection_check_strategy.iterate_samples(value)
         for i, element in enumerate(samples):
             try:
                 check_type_internal(element, element_type, memo)
@@ -369,88 +421,132 @@ def check_tuple(
 
 
 def check_union(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     errors: dict[str, TypeCheckError] = {}
-    for type_ in args:
-        try:
-            check_type_internal(value, type_, memo)
-            return
-        except TypeCheckError as exc:
-            errors[get_type_name(type_)] = exc
+    try:
+        for type_ in args:
+            try:
+                check_type_internal(value, type_, memo)
+                return
+            except TypeCheckError as exc:
+                errors[get_type_name(type_)] = exc
+
+        formatted_errors = indent(
+            "\n".join(f"{key}: {error}" for key, error in errors.items()), "  "
+        )
+    finally:
+        del errors  # avoid creating ref cycle
 
-    formatted_errors = indent(
-        "\n".join(f"{key}: {error}" for key, error in errors.items()), "  "
-    )
     raise TypeCheckError(f"did not match any element in the union:\n{formatted_errors}")
 
 
 def check_uniontype(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
+    if not args:
+        return check_instance(value, types.UnionType, (), memo)
+
     errors: dict[str, TypeCheckError] = {}
-    for type_ in args:
-        try:
-            check_type_internal(value, type_, memo)
-            return
-        except TypeCheckError as exc:
-            errors[get_type_name(type_)] = exc
+    try:
+        for type_ in args:
+            try:
+                check_type_internal(value, type_, memo)
+                return
+            except TypeCheckError as exc:
+                errors[get_type_name(type_)] = exc
+
+        formatted_errors = indent(
+            "\n".join(f"{key}: {error}" for key, error in errors.items()), "  "
+        )
+    finally:
+        del errors  # avoid creating ref cycle
 
-    formatted_errors = indent(
-        "\n".join(f"{key}: {error}" for key, error in errors.items()), "  "
-    )
     raise TypeCheckError(f"did not match any element in the union:\n{formatted_errors}")
 
 
 def check_class(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
-    if not isclass(value):
+    if not isclass(value) and not isinstance(value, generic_alias_types):
         raise TypeCheckError("is not a class")
 
-    # Needed on Python 3.7+
     if not args:
         return
 
-    expected_class = args[0]
+    if isinstance(args[0], ForwardRef):
+        expected_class = evaluate_forwardref(args[0], memo)
+    else:
+        expected_class = args[0]
+
+    if type(expected_class) in type_alias_types:
+        expected_class = expected_class.__value__
+
     if expected_class is Any:
         return
+    elif expected_class is typing_extensions.Self:
+        check_self(value, get_origin(expected_class), get_args(expected_class), memo)
     elif getattr(expected_class, "_is_protocol", False):
         check_protocol(value, expected_class, (), memo)
     elif isinstance(expected_class, TypeVar):
         check_typevar(value, expected_class, (), memo, subclass_check=True)
     elif get_origin(expected_class) is Union:
         errors: dict[str, TypeCheckError] = {}
-        for arg in get_args(expected_class):
-            if arg is Any:
-                return
+        try:
+            for arg in get_args(expected_class):
+                if arg is Any:
+                    return
 
-            try:
-                check_class(value, type, (arg,), memo)
-                return
-            except TypeCheckError as exc:
-                errors[get_type_name(arg)] = exc
-        else:
-            formatted_errors = indent(
-                "\n".join(f"{key}: {error}" for key, error in errors.items()), "  "
-            )
+                try:
+                    check_class(value, type, (arg,), memo)
+                    return
+                except TypeCheckError as exc:
+                    errors[get_type_name(arg)] = exc
+            else:
+                formatted_errors = indent(
+                    "\n".join(f"{key}: {error}" for key, error in errors.items()), "  "
+                )
+                raise TypeCheckError(
+                    f"did not match any element in the union:\n{formatted_errors}"
+                )
+        finally:
+            del errors  # avoid creating ref cycle
+    else:
+        if isinstance(expected_class, generic_alias_types):
+            expected_class = get_origin(expected_class)
+
+        if isinstance(value, generic_alias_types):
+            value = get_origin(value)
+
+        if not issubclass(value, expected_class):
             raise TypeCheckError(
-                f"did not match any element in the union:\n{formatted_errors}"
+                f"is not a subclass of {qualified_name(expected_class)}"
             )
-    elif not issubclass(value, expected_class):
-        raise TypeCheckError(f"is not a subclass of {qualified_name(expected_class)}")
 
 
 def check_newtype(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
-    supertype = origin_type.__supertype__
-    if not isinstance(value, supertype):
-        raise TypeCheckError(f"is not an instance of {qualified_name(supertype)}")
+    check_type_internal(value, origin_type.__supertype__, memo)
 
 
 def check_instance(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if not isinstance(value, origin_type):
         raise TypeCheckError(f"is not an instance of {qualified_name(origin_type)}")
@@ -466,7 +562,7 @@ def check_typevar(
 ) -> None:
     if origin_type.__bound__ is not None:
         annotation = (
-            Type[origin_type.__bound__] if subclass_check else origin_type.__bound__
+            type[origin_type.__bound__] if subclass_check else origin_type.__bound__
         )
         check_type_internal(value, annotation, memo)
     elif origin_type.__constraints__:
@@ -483,18 +579,24 @@ def check_typevar(
                 get_type_name(constraint) for constraint in origin_type.__constraints__
             )
             raise TypeCheckError(
-                f"does not match any of the constraints " f"({formatted_constraints})"
+                f"does not match any of the constraints ({formatted_constraints})"
             )
 
 
+def _is_literal_type(typ: object) -> bool:
+    return typ is typing.Literal or typ is typing_extensions.Literal
+
+
 def check_literal(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     def get_literal_args(literal_args: tuple[Any, ...]) -> tuple[Any, ...]:
         retval: list[Any] = []
         for arg in literal_args:
-            if get_origin(arg) is Literal:
-                # The first check works on py3.6 and lower, the second one on py3.7+
+            if _is_literal_type(get_origin(arg)):
                 retval.extend(get_literal_args(arg.__args__))
             elif arg is None or isinstance(arg, (int, str, bytes, bool, Enum)):
                 retval.append(arg)
@@ -519,19 +621,38 @@ def check_literal(
 
 
 def check_literal_string(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     check_type_internal(value, str, memo)
 
 
 def check_typeguard(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     check_type_internal(value, bool, memo)
 
 
+def check_none(
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
+) -> None:
+    if value is not None:
+        raise TypeCheckError("is not None")
+
+
 def check_number(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if origin_type is complex and not isinstance(value, (complex, float, int)):
         raise TypeCheckError("is neither complex, float or int")
@@ -540,7 +661,10 @@ def check_number(
 
 
 def check_io(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if origin_type is TextIO or (origin_type is IO and args == (str,)):
         if not isinstance(value, TextIOBase):
@@ -552,41 +676,235 @@ def check_io(
         raise TypeCheckError("is not an I/O object")
 
 
-def check_protocol(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
-) -> None:
-    # TODO: implement proper compatibility checking and support non-runtime protocols
-    if getattr(origin_type, "_is_runtime_protocol", False):
-        if not isinstance(value, origin_type):
-            raise TypeCheckError(
-                f"is not compatible with the {origin_type.__qualname__} protocol"
-            )
-    else:
-        warnings.warn(
-            f"Typeguard cannot check the {origin_type.__qualname__} protocol because "
-            f"it is a non-runtime protocol. If you would like to type check this "
-            f"protocol, please use @typing.runtime_checkable"
+def check_signature_compatible(subject: type, protocol: type, attrname: str) -> None:
+    subject_attr = getattr(subject, attrname)
+    try:
+        subject_sig = inspect.signature(subject_attr)
+    except ValueError:
+        return  # this can happen with builtins where the signature cannot be retrieved
+
+    protocol_sig = inspect.signature(getattr(protocol, attrname))
+    protocol_type: typing.Literal["instance", "class", "static"] = "instance"
+    subject_type: typing.Literal["instance", "class", "static"] = "instance"
+
+    # Check if the protocol-side method is a class method or static method
+    for klass in protocol.__mro__:
+        if attrname in klass.__dict__:
+            descriptor = klass.__dict__[attrname]
+            if isinstance(descriptor, staticmethod):
+                protocol_type = "static"
+            elif isinstance(descriptor, classmethod):
+                protocol_type = "class"
+
+            break
+
+    # Check if the subject-side method is a class method or static method
+    for klass in subject.__mro__:
+        if attrname in klass.__dict__:
+            descriptor = klass.__dict__[attrname]
+            if isinstance(descriptor, staticmethod):
+                subject_type = "static"
+            elif isinstance(descriptor, classmethod):
+                subject_type = "class"
+
+            break
+
+    if protocol_type == "instance" and subject_type != "instance":
+        raise TypeCheckError(
+            f"should be an instance method but it's a {subject_type} method"
         )
+    elif protocol_type != "instance" and subject_type == "instance":
+        raise TypeCheckError(
+            f"should be a {protocol_type} method but it's an instance method"
+        )
+
+    expected_varargs = any(
+        param
+        for param in protocol_sig.parameters.values()
+        if param.kind is Parameter.VAR_POSITIONAL
+    )
+    has_varargs = any(
+        param
+        for param in subject_sig.parameters.values()
+        if param.kind is Parameter.VAR_POSITIONAL
+    )
+    if expected_varargs and not has_varargs:
+        raise TypeCheckError("should accept variable positional arguments but doesn't")
+
+    protocol_has_varkwargs = any(
+        param
+        for param in protocol_sig.parameters.values()
+        if param.kind is Parameter.VAR_KEYWORD
+    )
+    subject_has_varkwargs = any(
+        param
+        for param in subject_sig.parameters.values()
+        if param.kind is Parameter.VAR_KEYWORD
+    )
+    if protocol_has_varkwargs and not subject_has_varkwargs:
+        raise TypeCheckError("should accept variable keyword arguments but doesn't")
+
+    # Check that the callable has at least the expect amount of positional-only
+    # arguments (and no extra positional-only arguments without default values)
+    if not has_varargs:
+        protocol_args = [
+            param
+            for param in protocol_sig.parameters.values()
+            if param.kind
+            in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
+        ]
+        subject_args = [
+            param
+            for param in subject_sig.parameters.values()
+            if param.kind
+            in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
+        ]
+
+        # Remove the "self" parameter from the protocol arguments to match
+        if protocol_type == "instance":
+            protocol_args.pop(0)
+
+        # Remove the "self" parameter from the subject arguments to match
+        if subject_type == "instance":
+            subject_args.pop(0)
+
+        for protocol_arg, subject_arg in zip_longest(protocol_args, subject_args):
+            if protocol_arg is None:
+                if subject_arg.default is Parameter.empty:
+                    raise TypeCheckError("has too many mandatory positional arguments")
+
+                break
+
+            if subject_arg is None:
+                raise TypeCheckError("has too few positional arguments")
+
+            if (
+                protocol_arg.kind is Parameter.POSITIONAL_OR_KEYWORD
+                and subject_arg.kind is Parameter.POSITIONAL_ONLY
+            ):
+                raise TypeCheckError(
+                    f"has an argument ({subject_arg.name}) that should not be "
+                    f"positional-only"
+                )
+
+            if (
+                protocol_arg.kind is Parameter.POSITIONAL_OR_KEYWORD
+                and protocol_arg.name != subject_arg.name
+            ):
+                raise TypeCheckError(
+                    f"has a positional argument ({subject_arg.name}) that should be "
+                    f"named {protocol_arg.name!r} at this position"
+                )
+
+    protocol_kwonlyargs = {
+        param.name: param
+        for param in protocol_sig.parameters.values()
+        if param.kind is Parameter.KEYWORD_ONLY
+    }
+    subject_kwonlyargs = {
+        param.name: param
+        for param in subject_sig.parameters.values()
+        if param.kind is Parameter.KEYWORD_ONLY
+    }
+    if not subject_has_varkwargs:
+        # Check that the signature has at least the required keyword-only arguments, and
+        # no extra mandatory keyword-only arguments
+        if missing_kwonlyargs := [
+            param.name
+            for param in protocol_kwonlyargs.values()
+            if param.name not in subject_kwonlyargs
+        ]:
+            raise TypeCheckError(
+                "is missing keyword-only arguments: " + ", ".join(missing_kwonlyargs)
+            )
+
+    if not protocol_has_varkwargs:
+        if extra_kwonlyargs := [
+            param.name
+            for param in subject_kwonlyargs.values()
+            if param.default is Parameter.empty
+            and param.name not in protocol_kwonlyargs
+        ]:
+            raise TypeCheckError(
+                "has mandatory keyword-only arguments not present in the protocol: "
+                + ", ".join(extra_kwonlyargs)
+            )
+
+
+def check_protocol(
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
+) -> None:
+    origin_annotations = typing.get_type_hints(origin_type)
+    for attrname in sorted(typing_extensions.get_protocol_members(origin_type)):
+        if (annotation := origin_annotations.get(attrname)) is not None:
+            try:
+                subject_member = getattr(value, attrname)
+            except AttributeError:
+                raise TypeCheckError(
+                    f"is not compatible with the {origin_type.__qualname__} "
+                    f"protocol because it has no attribute named {attrname!r}"
+                ) from None
+
+            try:
+                check_type_internal(subject_member, annotation, memo)
+            except TypeCheckError as exc:
+                raise TypeCheckError(
+                    f"is not compatible with the {origin_type.__qualname__} "
+                    f"protocol because its {attrname!r} attribute {exc}"
+                ) from None
+        elif callable(getattr(origin_type, attrname)):
+            try:
+                subject_member = getattr(value, attrname)
+            except AttributeError:
+                raise TypeCheckError(
+                    f"is not compatible with the {origin_type.__qualname__} "
+                    f"protocol because it has no method named {attrname!r}"
+                ) from None
+
+            if not callable(subject_member):
+                raise TypeCheckError(
+                    f"is not compatible with the {origin_type.__qualname__} "
+                    f"protocol because its {attrname!r} attribute is not a callable"
+                )
+
+            # TODO: implement assignability checks for parameter and return value
+            #  annotations
+            subject = value if isclass(value) else value.__class__
+            try:
+                check_signature_compatible(subject, origin_type, attrname)
+            except TypeCheckError as exc:
+                raise TypeCheckError(
+                    f"is not compatible with the {origin_type.__qualname__} "
+                    f"protocol because its {attrname!r} method {exc}"
+                ) from None
 
 
 def check_byteslike(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     if not isinstance(value, (bytearray, bytes, memoryview)):
         raise TypeCheckError("is not bytes-like")
 
 
 def check_self(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
-    if not isinstance(memo, CallMemo) or memo.self_type is None:
+    if memo.self_type is None:
         raise TypeCheckError("cannot be checked against Self outside of a method call")
 
     if isclass(value):
         if not issubclass(value, memo.self_type):
             raise TypeCheckError(
-                f"is not an instance of the self type "
-                f"({qualified_name(memo.self_type)})"
+                f"is not a subclass of the self type ({qualified_name(memo.self_type)})"
             )
     elif not isinstance(value, memo.self_type):
         raise TypeCheckError(
@@ -595,19 +913,19 @@ def check_self(
 
 
 def check_paramspec(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+    value: Any,
+    origin_type: Any,
+    args: tuple[Any, ...],
+    memo: TypeCheckMemo,
 ) -> None:
     pass  # No-op for now
 
 
-def check_instanceof(
-    value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo
+def check_type_internal(
+    value: Any,
+    annotation: Any,
+    memo: TypeCheckMemo,
 ) -> None:
-    if not isinstance(value, origin_type):
-        raise TypeCheckError(f"is not an instance of {qualified_name(origin_type)}")
-
-
-def check_type_internal(value: Any, annotation: Any, memo: TypeCheckMemo) -> None:
     """
     Check that the given object is compatible with the given type annotation.
 
@@ -624,16 +942,20 @@ def check_type_internal(value: Any, annotation: Any, memo: TypeCheckMemo) -> Non
         try:
             annotation = evaluate_forwardref(annotation, memo)
         except NameError:
-            if global_config.forward_ref_policy is ForwardRefPolicy.ERROR:
+            if memo.config.forward_ref_policy is ForwardRefPolicy.ERROR:
                 raise
-            elif global_config.forward_ref_policy is ForwardRefPolicy.WARN:
+            elif memo.config.forward_ref_policy is ForwardRefPolicy.WARN:
                 warnings.warn(
                     f"Cannot resolve forward reference {annotation.__forward_arg__!r}",
                     TypeHintWarning,
+                    stacklevel=get_stacklevel(),
                 )
 
             return
 
+    if type(annotation) in type_alias_types:
+        annotation = annotation.__value__
+
     if annotation is Any or annotation is SubclassableAny or isinstance(value, Mock):
         return
 
@@ -667,12 +989,22 @@ def check_type_internal(value: Any, annotation: Any, memo: TypeCheckMemo) -> Non
             checker(value, origin_type, args, memo)
             return
 
-    if not isinstance(value, origin_type):
-        raise TypeCheckError(f"is not an instance of {qualified_name(origin_type)}")
+    if isclass(origin_type):
+        if not isinstance(value, origin_type):
+            raise TypeCheckError(f"is not an instance of {qualified_name(origin_type)}")
+    elif type(origin_type) is str:  # noqa: E721
+        warnings.warn(
+            f"Skipping type check against {origin_type!r}; this looks like a "
+            f"string-form forward reference imported from another module",
+            TypeHintWarning,
+            stacklevel=get_stacklevel(),
+        )
 
 
 # Equality checks are applied to these
-origin_type_checkers = {
+origin_type_checkers: dict[
+    Any, Callable[[Any, Any, tuple[Any, ...], TypeCheckMemo], None]
+] = {
     bytes: check_byteslike,
     AbstractSet: check_set,
     BinaryIO: check_io,
@@ -682,16 +1014,16 @@ origin_type_checkers = {
     dict: check_mapping,
     Dict: check_mapping,
     float: check_number,
+    frozenset: check_set,
     IO: check_io,
     list: check_list,
     List: check_list,
-    Literal: check_literal,
-    LiteralString: check_literal_string,
+    typing.Literal: check_literal,
     Mapping: check_mapping,
     MutableMapping: check_mapping,
+    None: check_none,
     collections.abc.Mapping: check_mapping,
     collections.abc.MutableMapping: check_mapping,
-    Self: check_self,
     Sequence: check_sequence,
     collections.abc.Sequence: check_sequence,
     collections.abc.Set: check_set,
@@ -702,11 +1034,28 @@ origin_type_checkers = {
     Tuple: check_tuple,
     type: check_class,
     Type: check_class,
-    TypeGuard: check_typeguard,
     Union: check_union,
+    # On some versions of Python, these may simply be re-exports from "typing",
+    # but exactly which Python versions is subject to change.
+    # It's best to err on the safe side and just always specify these.
+    typing_extensions.Literal: check_literal,
+    typing_extensions.LiteralString: check_literal_string,
+    typing_extensions.Self: check_self,
+    typing_extensions.TypeGuard: check_typeguard,
 }
 if sys.version_info >= (3, 10):
     origin_type_checkers[types.UnionType] = check_uniontype
+    origin_type_checkers[typing.TypeGuard] = check_typeguard
+
+if sys.version_info >= (3, 11):
+    origin_type_checkers.update(
+        {typing.LiteralString: check_literal_string, typing.Self: check_self}
+    )
+
+if sys.version_info >= (3, 12):
+    type_alias_types = (typing_extensions.TypeAliasType, typing.TypeAliasType)
+else:
+    type_alias_types = (typing_extensions.TypeAliasType,)
 
 
 def builtin_checker_lookup(
@@ -718,7 +1067,8 @@ def builtin_checker_lookup(
     elif is_typeddict(origin_type):
         return check_typed_dict
     elif isclass(origin_type) and issubclass(
-        origin_type, Tuple  # type: ignore[arg-type]
+        origin_type,
+        Tuple,  # type: ignore[arg-type]
     ):
         # NamedTuple
         return check_tuple
@@ -737,7 +1087,7 @@ def builtin_checker_lookup(
         and getattr(origin_type, "__qualname__", "").startswith("NewType.")
         and hasattr(origin_type, "__supertype__")
     ):
-        # typing.NewType on Python 3.9 and below
+        # typing.NewType on Python 3.9
         return check_newtype
 
     return None
@@ -762,12 +1112,15 @@ def load_plugins() -> None:
             plugin = ep.load()
         except Exception as exc:
             warnings.warn(
-                f"Failed to load plugin {ep.name!r}: " f"{qualified_name(exc)}: {exc}"
+                f"Failed to load plugin {ep.name!r}: {qualified_name(exc)}: {exc}",
+                stacklevel=2,
             )
             continue
 
         if not callable(plugin):
-            warnings.warn(f"Plugin {ep} returned a non-callable object: {plugin!r}")
+            warnings.warn(
+                f"Plugin {ep} returned a non-callable object: {plugin!r}", stacklevel=2
+            )
             continue
 
         checker_lookup_functions.insert(0, plugin)
