@@ -1,54 +1,72 @@
 from __future__ import annotations
 
+import gc
 import inspect
 import sys
 from importlib import import_module
-from types import CodeType, FunctionType
-from typing import TYPE_CHECKING, Any, Callable, ForwardRef, Union
+from inspect import currentframe
+from types import CodeType, FrameType, FunctionType
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    ForwardRef,
+    Union,
+    cast,
+    final,
+    get_args,
+    get_origin,
+)
 from weakref import WeakValueDictionary
 
 if TYPE_CHECKING:
     from ._memo import TypeCheckMemo
 
-if sys.version_info >= (3, 10):
-    from typing import get_args, get_origin
+_functions_map: WeakValueDictionary[CodeType, FunctionType] = WeakValueDictionary()
+
+if sys.version_info >= (3, 14):
 
     def evaluate_forwardref(forwardref: ForwardRef, memo: TypeCheckMemo) -> Any:
-        return forwardref._evaluate(memo.globals, memo.locals, frozenset())
+        # If the ForwardRef has a module, try that module's namespace first.
+        # This is needed because Python 3.14's ForwardRef.evaluate() requires
+        # all referenced names to be available in the provided globals/locals.
+        if getattr(forwardref, "__forward_module__", None):
+            try:
+                # Not passing globals / locals defaults to those of the caller
+                return forwardref.evaluate(type_params=())
+            except NameError:
+                # Fall back to caller's namespace for backwards compatibility
+                pass
 
+        return forwardref.evaluate(
+            globals=memo.globals, locals=memo.locals, type_params=()
+        )
+elif sys.version_info >= (3, 13):
+
+    def evaluate_forwardref(forwardref: ForwardRef, memo: TypeCheckMemo) -> Any:
+        return forwardref._evaluate(
+            memo.globals, memo.locals, type_params=(), recursive_guard=frozenset()
+        )
 else:
-    from typing_extensions import get_args, get_origin
-
-    evaluate_extra_args: tuple[frozenset[Any], ...] = (
-        (frozenset(),) if sys.version_info >= (3, 9) else ()
-    )
 
     def evaluate_forwardref(forwardref: ForwardRef, memo: TypeCheckMemo) -> Any:
-        from ._union_transformer import compile_type_hint, type_substitutions
-
-        if not forwardref.__forward_evaluated__:
-            forwardref.__forward_code__ = compile_type_hint(forwardref.__forward_arg__)
-
         try:
-            return forwardref._evaluate(memo.globals, memo.locals, *evaluate_extra_args)
+            return forwardref._evaluate(
+                memo.globals, memo.locals, recursive_guard=frozenset()
+            )
         except NameError:
             if sys.version_info < (3, 10):
                 # Try again, with the type substitutions (list -> List etc.) in place
                 new_globals = memo.globals.copy()
                 new_globals.setdefault("Union", Union)
-                if sys.version_info < (3, 9):
-                    new_globals.update(type_substitutions)
 
                 return forwardref._evaluate(
-                    new_globals, memo.locals or new_globals, *evaluate_extra_args
+                    new_globals, memo.locals or new_globals, recursive_guard=frozenset()
                 )
 
             raise
 
 
-_functions_map: WeakValueDictionary[CodeType, FunctionType] = WeakValueDictionary()
-
-
 def get_type_name(type_: Any) -> str:
     name: str
     for attrname in "__name__", "_name", "__forward_arg__":
@@ -76,7 +94,11 @@ def get_type_name(type_: Any) -> str:
 
         name += f"[{formatted_args}]"
 
-    module = getattr(type_, "__module__", None)
+    # For ForwardRefs, use the module stored on the object if available
+    if hasattr(type_, "__forward_module__"):
+        module = type_.__forward_module__
+    else:
+        module = getattr(type_, "__module__", None)
     if module and module not in (None, "typing", "typing_extensions", "builtins"):
         name = module + "." + name
 
@@ -91,7 +113,9 @@ def qualified_name(obj: Any, *, add_class_prefix: bool = False) -> str:
     the module name stripped from the generated name.
 
     """
-    if inspect.isclass(obj):
+    if obj is None:
+        return "None"
+    elif inspect.isclass(obj):
         prefix = "class " if add_class_prefix else ""
         type_ = obj
     else:
@@ -137,3 +161,58 @@ def is_method_of(obj: object, cls: type) -> bool:
         and obj.__module__ == cls.__module__
         and obj.__qualname__.startswith(cls.__qualname__ + ".")
     )
+
+
+def get_stacklevel() -> int:
+    level = 1
+    frame = cast(FrameType, currentframe()).f_back
+    while frame and frame.f_globals.get("__name__", "").startswith("typeguard."):
+        level += 1
+        frame = frame.f_back
+
+    return level
+
+
+def find_function(frame: FrameType) -> Callable[..., Any]:
+    """
+    Return a function object from the garbage collector that matches the frame's code object.
+
+    This process is unreliable as several function objects could use the same code object.
+    Fortunately the likelihood of this happening with the combination of the function objects
+    having different type annotations is a very rare occurrence.
+
+    :param frame: a frame object
+    :return: a function object if one was found, ``None`` if not
+
+    """
+    func = _functions_map.get(frame.f_code)
+    if func is None:
+        for obj in gc.get_referrers(frame.f_code):
+            if inspect.isfunction(obj):
+                if func is None:
+                    # The first match was found
+                    func = obj
+                else:
+                    # A second match was found
+                    raise LookupError(
+                        "found more than one match when looking for the target function"
+                    )
+
+        # Cache the result for future lookups
+        if func is not None:
+            _functions_map[frame.f_code] = func
+        else:
+            raise LookupError("target function not found")
+
+    return func
+
+
+@final
+class Unset:
+    __slots__ = ()
+
+    def __repr__(self) -> str:
+        return "<unset>"
+
+
+unset = Unset()
