@@ -1,12 +1,15 @@
 import asyncio
+import subprocess
 import sys
 from contextlib import contextmanager
+from pathlib import Path
 from textwrap import dedent
 from typing import (
     Any,
     AsyncGenerator,
     AsyncIterable,
     AsyncIterator,
+    Dict,
     Generator,
     Iterable,
     Iterator,
@@ -16,7 +19,7 @@ from unittest.mock import Mock
 
 import pytest
 
-from typeguard import TypeCheckError, suppress_type_checks, typechecked
+from typeguard import TypeCheckError, typechecked
 
 if sys.version_info >= (3, 11):
     from typing import Self
@@ -369,7 +372,7 @@ class TestSelf:
             def method(self) -> Self:
                 return self
 
-        Foo().method
+        Foo().method()
 
     def test_return_invalid(self):
         class Foo:
@@ -453,6 +456,29 @@ class TestSelf:
             rf"test_classmethod_arg_invalid\.<locals>\.Foo\)"
         )
 
+    def test_self_type_valid(self):
+        class Foo:
+            @typechecked
+            def method(cls, subclass: type[Self]) -> None:
+                pass
+
+        class Bar(Foo):
+            pass
+
+        Foo().method(Bar)
+
+    def test_self_type_invalid(self):
+        class Foo:
+            @typechecked
+            def method(cls, subclass: type[Self]) -> None:
+                pass
+
+        pytest.raises(TypeCheckError, Foo().method, int).match(
+            rf'argument "subclass" \(class int\) is not a subclass of the self type '
+            rf"\({__name__}\.{self.__class__.__name__}\."
+            rf"test_self_type_invalid\.<locals>\.Foo\)"
+        )
+
 
 class TestMock:
     def test_mock_argument(self):
@@ -530,15 +556,6 @@ def test_retain_dunder_attributes():
     assert foo.__defaults__ == ("foo",)
 
 
-def test_suppressed_checking():
-    @typechecked
-    def foo(x: str) -> None:
-        pass
-
-    with suppress_type_checks():
-        foo(1)
-
-
 @pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires ast.unparse()")
 def test_debug_instrumentation(monkeypatch, capsys):
     monkeypatch.setattr("typeguard.config.debug_instrumentation", True)
@@ -553,11 +570,14 @@ def test_debug_instrumentation(monkeypatch, capsys):
         Source code of test_debug_instrumentation.<locals>.foo() after instrumentation:
         ----------------------------------------------
         def foo(a: str) -> int:
-            from typeguard import CallMemo
-            from typeguard._functions import check_argument_types, check_return_type
-            call_memo = CallMemo(foo, locals())
-            check_argument_types(call_memo)
-            return check_return_type(6, call_memo)
+            from typeguard import TypeCheckMemo
+            from typeguard._functions import check_argument_types_internal, \
+check_return_type_internal
+            memo = TypeCheckMemo(globals(), locals())
+            check_argument_types_internal('test_debug_instrumentation.<locals>.foo', \
+{'a': (a, str)}, memo)
+            return check_return_type_internal('test_debug_instrumentation.<locals>.foo', \
+6, int, memo)
         ----------------------------------------------
         """
     )
@@ -592,3 +612,99 @@ def test_existing_method_decorator():
 
     with Foo().method(6) as value:
         assert value == 7
+
+
+@pytest.mark.parametrize(
+    "flags, expected_return_code",
+    [
+        pytest.param([], 1, id="debug"),
+        pytest.param(["-O"], 0, id="O"),
+        pytest.param(["-OO"], 0, id="OO"),
+    ],
+)
+def test_typechecked_disabled_in_optimized_mode(
+    tmp_path: Path, flags: List[str], expected_return_code: int
+):
+    code = dedent(
+        """
+        from typeguard import typechecked
+
+        @typechecked
+        def foo(x: int) -> None:
+            pass
+
+        foo("a")
+        """
+    )
+    script_path = tmp_path / "code.py"
+    script_path.write_text(code)
+    process = subprocess.run(
+        [sys.executable, *flags, str(script_path)], capture_output=True
+    )
+    assert process.returncode == expected_return_code
+    if process.returncode == 1:
+        assert process.stderr.strip().endswith(
+            b'typeguard.TypeCheckError: argument "x" (str) is not an instance of int'
+        )
+
+
+def test_reference_imported_name_from_method() -> None:
+    # Regression test for #362
+    @typechecked
+    class A:
+        def foo(self) -> Dict[str, Any]:
+            return {}
+
+    A().foo()
+
+
+def test_getter_setter():
+    """Regression test for #355."""
+
+    @typechecked
+    class Foo:
+        def __init__(self, x: int):
+            self._x = x
+
+        @property
+        def x(self) -> int:
+            return self._x
+
+        @x.setter
+        def x(self, value: int) -> None:
+            self._x = value
+
+    f = Foo(1)
+    f.x = 2
+    assert f.x == 2
+    with pytest.raises(TypeCheckError):
+        f.x = "foo"
+
+
+def test_duplicate_method():
+    class Foo:
+        def x(self) -> str:
+            return "first"
+
+        @typechecked()
+        def x(self, value: int) -> str:  # noqa: F811
+            return "second"
+
+    assert Foo().x(1) == "second"
+    with pytest.raises(TypeCheckError):
+        Foo().x("wrong")
+
+
+def test_duplicate_function():
+    @typechecked
+    def foo() -> list[int]:  # noqa: F811
+        return [x for x in range(5)]
+
+    foo1 = foo
+
+    @typechecked
+    def foo() -> list[int]:  # noqa: F811
+        return [x for x in range(5, 10)]
+
+    assert foo1() == [0, 1, 2, 3, 4]
+    assert foo() == [5, 6, 7, 8, 9]
